import base64
import os
from typing import Annotated

import click
import pulumi_ovh as ovh
from pulumi import CustomTimeouts, ResourceOptions

from .. import JSON, DefaultOpt, data
from .base import StackName, default, defaults

DEFAULTS = {
    "instance_opts": ("OVH_INSTANCE_OPTS", dict()),
}


def find_resource_id(items, name: str, resource_type: str, region: str) -> str:
    """Find a resource ID by name and region."""
    resource_id = next(
        (item.id for item in items if item.name == name and item.region == region),
        None,
    )
    if not resource_id:
        raise ValueError(
            f"The `{name}` {resource_type} is not supported in the `{region}` region"
        )
    return resource_id


def resources_ovh(
    project_id: Annotated[
        str,
        DefaultOpt(
            ["--project-id"],
            type=str,
            help="OVHcloud project UUID",
            required=True,
            # don't use envvar Click parameter here for the default value
            # as it's not picked up when using the runner.create() instead of the CLI
        ),
    ] = os.environ.get("OVH_CLOUD_PROJECT_SERVICE"),
    region: Annotated[
        str,
        DefaultOpt(["--region"], type=click.Choice(data.regions("ovh")), help="Region"),
        StackName(),
    ] = os.environ.get("OVH_REGION", "EU-WEST-PAR"),
    instance: Annotated[
        str,
        DefaultOpt(
            ["--instance"],
            type=click.Choice(data.servers("ovh")),
            help="Instance type",
        ),
        StackName(),
    ] = os.environ.get("INSTANCE_TYPE", "c3-4"),
    image_name: Annotated[
        str, DefaultOpt(["--image-name"], type=str, help="Boot image name")
    ] = "Ubuntu 25.04",
    instance_opts: Annotated[
        str,
        DefaultOpt(
            ["--instance-opts"],
            type=JSON,
            default=defaults(DEFAULTS, "instance_opts"),
            help="Pulumi ovh.cloudproject.Instance options",
        ),
    ] = default(DEFAULTS, "instance_opts"),
    user_data: Annotated[
        str | None,
        DefaultOpt(
            ["--user-data"],
            type=str,
            help="Base64 encoded string with user_data script to run at boot",
        ),
    ] = os.environ.get("USER_DATA", None),
):
    """Define an OVH cloud instance.

    Required environment variables/configuration:

    - `OVH_CLOUD_PROJECT_SERVICE` (OVHcloud project UUID)
    - `OVH_ENDPOINT` (e.g. ovh-eu)
    - `OVH_CLIENT_ID`
    - `OVH_CLIENT_SECRET`

    Required permissions:

    - `publicCloudProject:apiovh:operation/get`
    - `publicCloudProject:apiovh:flavor/get`
    - `publicCloudProject:apiovh:image/get`
    - `publicCloudProject:apiovh:instance/create`
    - `publicCloudProject:apiovh:region/instance/create`
    - `publicCloudProject:apiovh:region/instance/get`
    - `publicCloudProject:apiovh:instance/delete`

    Required scopes:

    - OVH account
    - OVH project
    """
    if user_data:
        instance_opts["user_data"] = base64.b64encode(user_data.encode()).decode()

    flavors = ovh.cloudproject.get_flavors(service_name=project_id)
    flavor_id = find_resource_id(flavors.flavors, instance, "instance type", region)

    images = ovh.cloudproject.get_images(service_name=project_id)
    image_id = find_resource_id(images.images, image_name, "image", region)

    ovh.cloudproject.Instance(
        instance,
        name=instance,
        boot_from={"image_id": image_id},
        flavor={"flavor_id": flavor_id},
        network={"public": True},
        billing_period="hourly",
        region=region,
        # OVH default timeout is 1 hour that is overkill, so decrease to 10 mins (like AWS default)
        opts=ResourceOptions(custom_timeouts=CustomTimeouts(create="10m")),
        **instance_opts,
    )



import inspect
import sys

from .alicloud import resources_alicloud
from .aws import resources_aws
from .azure import resources_azure
from .base import *
from .gcp import resources_gcp
from .hcloud import resources_hcloud
from .ovh import resources_ovh
from .upcloud import resources_upcloud

# method name prefix for initializing vendor-specific resources
PREFIX = "resources_"


supported_vendors = {
    name[len(PREFIX) :]
    for name, _ in inspect.getmembers(sys.modules[__name__], inspect.isfunction)
    if name.startswith(PREFIX)
}


__all__ = [
    "resources_alicloud",
    "resources_aws",
    "resources_azure",
    "resources_gcp",
    "resources_hcloud",
    "resources_ovh",
    "resources_upcloud",
    "supported_vendors",
]



from sc_crawler.tables import Server, ServerPrice, Region, Vendor, Zone
from sqlmodel import create_engine, Session, select
import sc_data


session = Session(create_engine(f"sqlite:///{sc_data.db.path}"))


def vendors():
    return session.exec(select(Vendor.vendor_id)).all()


def regions(vendor: str):
    return session.exec(select(Region.api_reference).where(Region.vendor_id == vendor)).all()


def zones(vendor: str):
    return session.exec(select(Zone.api_reference).where(Zone.vendor_id == vendor)).all()


def servers(vendor: str, region: str | None = None, zone: str | None = None):
    stmt = select(ServerPrice.server_id, Server.api_reference).join(Zone).join(Server).where(ServerPrice.vendor_id == vendor)
    if region:
        stmt = stmt.where(ServerPrice.region_id == region)
    if zone:
        stmt = stmt.where(ServerPrice.zone_id == zone)
    return [i[1] for i in session.exec(stmt.distinct()).all()]


def servers_vendors(vendor: str, region: str | None = None, zone: str | None = None):
    stmt = select(ServerPrice.vendor_id, ServerPrice.region_id, Zone.api_reference, ServerPrice.server_id).join(Zone).where(ServerPrice.vendor_id == vendor)
    if region:
        stmt = stmt.where(ServerPrice.region_id == region)
    if zone:
        stmt = stmt.where(ServerPrice.zone_id == zone)
    return session.exec(stmt.distinct()).all()


def server_cpu_architecture(vendor: str, server: str) -> str:
    return session.exec(select(Server.cpu_architecture).where(Server.vendor_id == vendor).where(Server.api_reference == server)).one().value


import copy
import os
from typing import Annotated

import click
import pulumi_upcloud as upcloud

from .. import JSON, DefaultOpt, data
from .base import StackName, default, defaults

DEFAULTS = {
    "instance_opts": (
        "UPCLOUD_INSTANCE_OPTS",
        dict(
            labels={"created-by": "sc-runner"},
            # tags=["sc-runner"],  # only main account can set tags
        ),
    ),
}


def resources_upcloud(
    region: Annotated[
        str,
        DefaultOpt(
            ["--region"], type=click.Choice(data.regions("upcloud")), help="Region"
        ),
        StackName(),
    ] = os.environ.get("UPCLOUD_REGION", "fi-hel1"),
    instance: Annotated[
        str,
        DefaultOpt(
            ["--instance"],
            type=click.Choice(data.servers("upcloud")),
            help="Server plan",
        ),
        StackName(),
    ] = os.environ.get("INSTANCE_TYPE", "DEV-1xCPU-4GB"),
    public_key: Annotated[
        str,
        DefaultOpt(["--public-key"], type=str, help="SSH public key for the root user"),
    ] = os.environ.get("SSH_PUBLIC_KEY", ""),
    user_data: Annotated[
        str | None,
        DefaultOpt(
            ["--user-data"],
            type=str,
            help="Base64 encoded string with user_data script to run at boot",
        ),
    ] = os.environ.get("USER_DATA", None),
    disk_size: Annotated[
        int, DefaultOpt(["--disk-size"], type=int, help="Boot disk size in GiBs")
    ] = int(os.environ.get("DISK_SIZE", 30)),
    instance_opts: Annotated[
        str,
        DefaultOpt(
            ["--instance-opts"],
            type=JSON,
            default=defaults(DEFAULTS, "instance_opts"),
            help="pulumi_upcloud.Server options",
        ),
    ] = default(DEFAULTS, "instance_opts"),
):
    # we don't want to modify the default
    instance_opts = copy.deepcopy(instance_opts)
    upcloud.Server(
        instance,
        hostname=instance.lower(),
        plan=instance,
        zone=region,
        login={"user": "root", "keys": [public_key], "create_password": False},
        template={
            "size": disk_size,
            "storage": "Ubuntu Server 24.04 LTS (Noble Numbat)",
        },
        network_interfaces=[{"type": "public"}],
        metadata=True,
        user_data=user_data,
        **instance_opts,
    )



import click
import json


class DefaultOpt(click.Option):
    def __init__(self, *args, **kwargs):
        kwargs["show_default"] = True
        super().__init__(*args, **kwargs)


class JsonParamType(click.ParamType):
    name = "json"

    def convert(self, value, param, ctx):
        try:
            return json.loads(value)
        except ValueError:
            self.fail(f"{value!r} is not a valid JSON!", param, ctx)


JSON = JsonParamType()